/************************************************
**** FIBHEAP:                                ****
**** classe che implementa gli heap          ****
**** di Fibonacci                            ****
****                                         ****
**** Larcheri Paolo       52 SI              ****
************************************************/

#include "FibHeap.h"
#include "ProgressBar.h"
#ifndef IO
 #define IO
 #include <iostream.h>
#endif

/***************************** COSTRUTTORE e DISTRUTTORE ******************************/

FibHeap::FibHeap() {
  /* inizializzazione degli attributi */
  this->testa = NULL;
  this->min = NULL;
  this->num_nodi = 0;
  this->num_alberi = 0;
}

FibHeap::~FibHeap() {
}

/********************************* METODI DI UTILITA' *********************************/
/*********************** settano o ritornano i vari attributi *************************/

Elem* FibHeap::getTesta() {
  return this->testa;
}

int FibHeap::getNum_nodi() {
  return this->num_nodi;
}

int FibHeap::getNum_alberi() {
  return this->num_alberi;
}

int FibHeap::getNum_marcati() {
  return this->num_marcati;
}

void FibHeap::setTesta(Elem *t) {
  this->testa = t;
}

void FibHeap::FibHeap::setMin(Elem *m) {
  this->testa = m;
}

void FibHeap::setNum_nodi(int nn) {
  this->num_nodi = nn;
}

void FibHeap::setNum_alberi(int na) {
  this->num_alberi = na;
}

void FibHeap::setNum_marcati(int nm) {
  this->num_marcati = nm;
}


/*********************** INSERISCE UN NUOVO ELEMENTO NELLO HEAP ***********************/
/* prende in input un numero e aggiunge allo heap un nuovo elemento con quella chiave */

void FibHeap::insert(long int chiave) {
  Elem *newElem = new Elem(chiave);
  
  /* aggiungo l'elemento in testa alla lista delle radici */
  if((this->testa) != NULL) {
    /* se lo heap non e' vuoto... */
    newElem->setDestro(this->testa);
    newElem->setSinistro((this->testa)->getSinistro());
    (testa->getSinistro())->setDestro(newElem);
    (this->testa)->setSinistro(newElem);
    
  }
  else {
    /* se lo heap e' vuoto... */
    newElem->setDestro(newElem);
    newElem->setSinistro(newElem);
  }
  testa = newElem;
  
  /* eventualmente aggiorno il 'puntatore al minimo' dello heap */
  if ((this->min == NULL) || (min->getChiave() >= newElem->getChiave()))
    (this->min) = newElem;
  
  /* incremento il numero di nodi e di radici */  
  (this->num_alberi)++;
  (this->num_nodi)++;
}


/********************** RESTITUISCE L'ELEMENTO CON CHIAVE MINIMA *************************/

Elem* FibHeap::getMin() {
  return this->min;
}



/************************** UNIONE DI DUE HEAP DI FIBONACCI ******************************/
/* mette in coda allo heap le radici di un secondo heap del quale e' passato l'indirizzo */

void FibHeap::unify(FibHeap **FH2) {
  Elem *sin1, *sin2;
  
  /* se lo heap e' vouto, questo diventa praticamente quello di cui 
     e' passato l'indirizzo */
  if (this->getTesta() == NULL) {
    this->testa = (*FH2)->getTesta();
    this->min = (*FH2)->getMin();
    delete(*FH2);
    (*FH2) = new FibHeap();
    return;
  }
  /* se lo heap  di cui e' passato l'indirizzo e' vuoto, ritorno*/
  if ((*FH2)->getTesta() == NULL) {
    return;
  }
  
  /* appendo le radici di uno alla lista delle radici dell'altro */
  sin1 = (this->testa)->getSinistro();
  sin2 = ((*FH2)->getTesta())->getSinistro();
  sin1->setDestro((*FH2)->getTesta());
  sin2->setDestro(this->testa);
  ((*FH2)->getTesta())->setSinistro(sin1);
  (this->testa)->setSinistro(sin2);
  
  /* aggiorno gli attributi/contatori */
  this->setNum_nodi(this->getNum_nodi() + (*FH2)->getNum_nodi());  
  this->setNum_alberi(this->getNum_alberi() + (*FH2)->getNum_alberi());  
  this->setNum_marcati(this->getNum_marcati() + (*FH2)->getNum_marcati());  
  
  /* se serve aggiorno il puntatore al minimo */
  if((this->getMin())->getChiave()  >  ((*FH2)->getMin())->getChiave())
     (this->min) = ((*FH2)->getMin());
  
  /* libero la memoria occupata dallo heap passato per indirizzo
     e ne alloco dell'altra */   
  delete(*FH2);
  *FH2 = new FibHeap();
}


/********************** ESTRAZIONE DEL NODO CON CHIAVE MINIMA **************************/

Elem* FibHeap::extractMin() {
  Elem *figlio, *ultimo_figlio, *minimo;
  int num_figli;
  
  minimo = NULL;
    
  if((this->min) != NULL) {
    minimo = this->min;
    num_figli = minimo->getGrado();

   /* aggiungo alla lista delle radici tutti i figli di this->min */
    if((figlio = minimo->getFigli()) != NULL) {
      ultimo_figlio = figlio->getSinistro();
      /* tutti gli ex-figli di this->min non hanno piu' padre */
      do {
        figlio->setPadre(NULL);
	if (figlio->isMarked())
	  (this->num_marcati)--;
	figlio->unmark();
        figlio = figlio->getDestro();    
      } while(figlio != (ultimo_figlio->getDestro()));

      ultimo_figlio = figlio->getSinistro();
      figlio->setSinistro(minimo);
      ultimo_figlio->setDestro(minimo->getDestro());
      (minimo->getDestro())->setSinistro(ultimo_figlio);
      minimo->setDestro(figlio);
     
    }
    
    /* aggiorno gli attributi/contatori */
    (this->num_nodi)--;
    (this->num_alberi) += (num_figli-1);
        
    /* tolgo il minimo dalla lista della matrici */
    if((this->num_nodi) > 0) {
      (minimo->getSinistro())->setDestro(minimo->getDestro());
      (minimo->getDestro())->setSinistro(minimo->getSinistro());
    }
 
    /* aggiorno il minimo e la testa dello heap */
    if((minimo->getDestro()) == minimo) {
      this->min = NULL;
      this->testa = NULL;
    }
    else {
      if(this->testa == this->min)
        this->testa = minimo->getDestro();
      this->min = minimo->getDestro();
      this->consolidate();    /* RIMETTE A POSTO LO HEAP */
    }
  }
  
  /* ritorno il nodo con chiave minima */
  /* NOTA: sara' il chiamante a dover deallocare la memoria da lui occupata */
  return minimo;
}


/******************************* RIMETTE A POSTO LO HEAP **********************************/
/********************** - funzione di supporto della extractMin() - ***********************/

void FibHeap::consolidate() {
  /* NB: se A[i]=y allora y e' una radice con grado i */
  Elem **A, *nodo_curr1, *nodo_curr2, *deposito, *next;
  int i, degree, aut1aut2, fine;
  
  /* inizializzo A[] a NULL */
  A = (Elem**)malloc(((this->num_nodi)+1)*sizeof(Elem*));
  for(i=0; i<(this->num_nodi); i++) 
    A[i] = NULL;
  
  next = this->min;
  fine = this->num_alberi;
  
  /* per ogni radice dello heap... */
  for (i=1; i<=fine; i++) {  
    nodo_curr1 = next;
    next = next->getDestro();
    degree = nodo_curr1->getGrado();
    /* finche' c'e' un altra radice con lo stesso grado... */
    while(A[degree] != NULL) {
      nodo_curr2 = A[degree];
      if ((nodo_curr1->getChiave()) > (nodo_curr2->getChiave())) {
	deposito = nodo_curr1;
        nodo_curr1 = nodo_curr2;
        nodo_curr2 = deposito;
      }
      /* fondo i due alberi in unico albero di grado (degree+1) */
      this->link(nodo_curr2, nodo_curr1);
      A[degree] = NULL;
      degree++; 
    }
    
    A[degree] = nodo_curr1;
 }
  
  /* metto MOMENTANEAMENTE a NULL la testa e il minimo */
  (this->testa) = NULL;
  (this->min) = NULL;
  
  /* ricompongo la lista delle radici dello heap con tutti gli elementi contenuti in A[] */
  for (i=0; i<(this->num_nodi); i++) {
    if(A[i] != NULL) {
      if((this->testa) == NULL) {
        A[i]->setDestro(A[i]);
        A[i]->setSinistro(A[i]);
      }
      else {
     	A[i]->setSinistro((this->testa)->getSinistro());
	((this->testa)->getSinistro())->setDestro(A[i]);
	(this->testa)->setSinistro(A[i]);
        A[i]->setDestro(this->testa);
      }
      
      (this->testa) = A[i];
      
      /* se necessario aggiorno il minimo */
      if((this->min == NULL) || ((this->min)->getChiave() > A[i]->getChiave())) 
        this->min = A[i];
    }
  }
  
  /* libero la a cui punta A[]*/
  free(A);
}

/****************************** FONDE DUE ALBERI IN UNO ***********************************/
/********************** - funzione di supporto della extractMin() - ***********************/

void FibHeap::link(Elem *T1, Elem *T2) {
  Elem *figlio;
  
  /* Tolgo T1 dalla lista delle radici */
  (T1->getSinistro())->setDestro(T1->getDestro());
  (T1->getDestro())->setSinistro(T1->getSinistro());

  /* T1 diventa PRIMO figlio di T2 */  
  T1->setPadre(T2);
  T2->setGrado(T2->getGrado() + 1);
  this->setNum_alberi(this->getNum_alberi() - 1);
  if((figlio=T2->getFigli()) != NULL) {
    T1->setDestro(figlio);
    T1->setSinistro(figlio->getSinistro());
    (figlio->getSinistro())->setDestro(T1);
    figlio->setSinistro(T1);
  }
  else{
    T1->setDestro(T1);
    T1->setSinistro(T1);
  }
  T2->setFigli(T1);
}

/*************** DECREMENTA LA CHIAVE DI UN ELEMENTO DATA LA SUA POSIZIONE *****************/

int FibHeap::decreaseKey(Elem* nodo, long int new_key) {
  Elem* pa;
  
  /* controllo il valore della nuova chiave */
  if(nodo->getChiave() < new_key) {
    cout << endl << "ATTENZIONE: la nuova chiave e' maggiore della chiave corrente!" << endl;
    return (0);
  }
  
  nodo->setChiave(new_key);
  pa = nodo->getPadre();
  
  if((pa != NULL)  &&  (pa->getChiave() > nodo->getChiave())) {
    /* se il nodo a cui e' stata decrementata la chiave ha padre si prosegue 
       a risistemare lo heap */
    this->cut(nodo, pa);
    this->cascadingCut(pa);
  }
  
  /* se necessario aggiorno il minimo */
  if((this->min->getChiave()) > (nodo->getChiave()))
    (this->min) = nodo;

  return (1);
}

/************* TAGLIA UN NODO DA UN ALBERO E LO METTE NELLA LISTA DELLE RADICI **************/
/*********************** - funzione di supporto alla decreaseKey() - ************************/

void FibHeap::cut(Elem* fi, Elem* pa) {
  
  /* tolgo il figlio dalla lista dei figli */
  if (pa->getGrado() == 1) 
    pa->setFigli(NULL);
  else {
    pa->setFigli(fi->getDestro());
    (fi->getDestro())->setSinistro(fi->getSinistro());
    (fi->getSinistro())->setDestro(fi->getDestro());
  }
  
  /* il figlio non ha padre e il padre ha un figlio in meno */
  fi->setPadre(NULL);
  pa->setGrado(pa->getGrado() - 1);
  
  /* aggiungo il figlio in testa alla lista delle radici */
  fi->setDestro(this->testa);
  fi->setSinistro(this->testa->getSinistro());
  (this->testa->getSinistro())->setDestro(fi);
  (this->testa)->setSinistro(fi);
  this->testa = fi; 
}

/************* TAGLIA UN NODO DA UN ALBERO E LO METTE NELLA LISTA DELLE RADICI **************/
/*********************** - funzione di supporto alla decreaseKey() - ************************/

void FibHeap::cascadingCut(Elem *curr_nodo){
  Elem *padre;
  
  padre = curr_nodo->getPadre();
  /* tramite chiamate ricorsive risalgo l'albero finche' non trovo un padre non marcato,
     altrimenti lo promuovo a radice e proseguo nella risalita */
  if(padre != NULL) {
    if(!(curr_nodo->isMarked()))
      curr_nodo->mark();
    else {
      this->cut(curr_nodo, padre);   
      this->cascadingCut(padre);
    }
  }
}

/******************************* SVUOTA LO HEAP DI FIBONACCI ************************************/

void FibHeap::reset() {
  Elem *minimo;
  ProgressBar pb = ProgressBar("DEALLOCAZIONE MEMORIA:", this->getNum_nodi(), 12, 40,'#');
  
  while(this->testa != NULL){
    /* per ogni nodo dello heap viene liberata la memoria */
    minimo = this->extractMin();
    delete(minimo);
    pb.Advance();
  }
}
