/************************************************
**** HEAPTEST:                               ****
**** programma urilizzato per testare        ****
**** le funzionalita' degli heap             ****
**** di Fibonacci                            ****
****                                         ****
**** Larcheri Paolo       52 SI              ****
************************************************/

#include "Menu.h"
#include "Elem.h"
#include "FibHeap.h"
#include "ProgressBar.h"
#include <unistd.h>
#include <fcntl.h>       
#include <sys/stat.h>    
#include <sys/types.h>  
#include <iostream.h>

#define MAXSIZE 50

FibHeap *FH1 = new FibHeap();
FibHeap *FH2 = new FibHeap();
char buff[MAXSIZE];

void setMenu(Menu*);
int gestioneScelta(int);
void gestioneINS();
void gestioneMIN();
void gestioneCANCMIN();
void gestioneRESET();
void gestioneLOAD();
void gestioneSAVE();
void gestioneUNION();

/***************************** PROGRAMMA PRINCIPALE ******************************/
/*stampa il menu finche' non viene scelta l'opzione "uscita"*/

void main() {
 int scelta, uscita = 0;
 Menu *mainMenu = new Menu("OPERAZIONI SULLO HEAP:");
 
 setMenu(mainMenu);  
 while(!uscita) {
   scelta = mainMenu->select();   
   uscita = gestioneScelta(scelta);
 }

}

/*************************** STAMPA A VIDEO DEL MENU ******************************/
/*Inserisce le voci nel menu*/

void setMenu(Menu* m) {
 m->addOpt("INSERISCI NUOVO ELEMENTO");
 m->addOpt("TROVA MINIMO E STAMPA INFO");
 m->addOpt("TROVA E CANCELLA MINIMO");
 m->addOpt("SVUOTA HEAP");
 m->addOpt("CARICA DA FILE");
 m->addOpt("SALVA SU FILE");
 m->addOpt("FONDI CON...");
 m->addOpt("------- USCITA ---------->");
}

/****************************** GESTIONE DELLA SCELTA *****************************/
/* Gestisce le chiamate alle procedure in base alla scelta dal menu */

int gestioneScelta(int sc) {
 int uscita = 0;
  
 switch (sc) {
   case 1:
     gestioneINS();
     break;
   case 2:
     gestioneMIN();
     break;
   case 3:
     gestioneCANCMIN();
     break;
   case 4:
     gestioneRESET();
     break;
   case 5:
     gestioneLOAD();
     break;
   case 6:
     gestioneSAVE();
     break;
   case 7:
     gestioneUNION();
     break;
   case 8:
     uscita = 1;
     break;
   default:
     cout << ":-( Scelta non valida!" << endl;
 }
 return uscita;
}

/*********************** GESTIONE INSERIMENTO NELLO HEAP ***************************/
/* Inserisce in FH1 un nuovo elemento di e' specificata la chiave */

void gestioneINS() {
 int ok = 0;
 long int chiave;
 char **endptr = (char**)malloc(sizeof(char*));
 
 while (!ok) {
   cout << endl << "--> Inserisci il valore del nuovo elemento (q per uscire): ";
   cin.getline(buff, MAXSIZE);
   
   switch (buff[0]) {
     case '\0':
       /* Senza questo controllo dando solo RETURN, la strtol() ritorna 0 */
       cout << ":-( Scelta non valida: inserire solo caratteri numerici!" << endl;
       cout.flush();
       sleep(1);
       break;
      
     case 'Q': case 'q':
       /* Opzione di uscita */
       ok = 1;
       break;
     
     default: 
       /* ATTENZIONE: la strtol() non ritorna errore se il numero e' troppo grande
        *             o troppo piccolo ma converte la stringa in MAXINT/MININT.
	*/
       chiave = strtol(buff, endptr, 10);
       if (**endptr == '\0') {
         /* Conversione a intero corretta! */
         FH1->insert(chiave);
	 cout << ":-) Elemento con valore " << chiave << " inserito correttamente!";
	 cout << endl;
	}
       else {
         /* Errore di conversione a intero! */
	 cout << ":-( Scelta non valida: inserire solo caratteri numerici!" << endl;
       }	  
   }  
 }
}

/**************************** GESTIONE RITORNO DEL MINIMO ******************************/
/* Stampa a video: valore del minimo, numero-nodi, num-radici, 
 *                 valore-radici e loro-grado 
 */
void gestioneMIN() {
 Elem *minimo, *nodo_curr;
 char riga[100];
 
 minimo = FH1->getMin();
 if (minimo == NULL) {
   cout << endl << ":-) MINIMO     = " << "<nessuno>";
   cout << endl << "    NUM NODI   = " << FH1->getNum_nodi();
   cout << endl << "    NUM RADICI = " << FH1->getNum_alberi() << endl;
   cout << "    RADICI : " << "    <nessuna>" << endl;
 }
 else {
   cout << endl << ":-) MINIMO     = " << minimo->getChiave();
   cout << endl << "    NUM NODI   = " << FH1->getNum_nodi();
   cout << endl << "    NUM RADICI = " << FH1->getNum_alberi();
   cout << endl << "    RADICI" << endl; 
   nodo_curr = FH1->getTesta();
   do {
     sprintf(riga, "       valore:%11d  |    grado:%4d", nodo_curr->getChiave(), nodo_curr->getGrado());
     cout << riga << endl;
     nodo_curr = nodo_curr->getDestro();  
   } while (nodo_curr != FH1->getTesta());
 }
}

/******************** GESTIONE RITORNO E CANCELLAZIONE DEL MINIMO *********************/
/* Estra il minimo da FH1 e ne libera la memoria (+ stampa info) */

void gestioneCANCMIN() {
 Elem* minimo;
 
 if(FH1->getMin() != NULL){
  /* Stampa info di FH1 prima della cancellazione del minimo */
   cout << endl << "************** PRIMA DELL'ESTRAZIONE DEL MINIMO ************"; 
   gestioneMIN();
  /* Estrazione del minimo */ 
   minimo = FH1->extractMin();
  /* Stampa info di FH1 dopo la cancellazione del minimo */
   cout << ":-) L'elemento minimo ha chiave " << minimo->getChiave();
   cout << " ed e' stato cancellato." << endl;
   cout << endl << "**************** DOPO L'ESTRAZIONE DEL MINIMO *************"; 
   gestioneMIN();
  /* Libera la memoria occupata dal nodo con chiave minima */ 
   delete(minimo);
 }
 else {
   cout << endl << "******************* LO HEAP E' VUOTO *********************"; 
   gestioneMIN();
 }
}

/************************* GESTIONE SVUTAMENTO DELLO HEAP *****************************/
/* Svuota lo heap -> libera la memoria occupata da ogni elemento */

void gestioneRESET() {
 if((FH1->getNum_nodi()) == 0) {
   cout << endl << ":-) Lo heap e' gia' vuoto!" << endl;
   return;
 }
 FH1->reset();
 cout << endl << ":-) Lo heap e' stato svuotato." << endl;
}

/*************************** GESTIONE CARICAMENTO DA FILE *****************************/
/* Carica fa file le chiavi del nuovo heap -> lo heap viene svuotato dei vecchi dati */

void gestioneLOAD() {
 int continua = 1;
 FILE *fptr;
 long int chiave;
 char **endptr = (char**)malloc(sizeof(char*));
 char *num = (char*)malloc(MAXSIZE*sizeof(char));
 /* Stampa warning */ 
 cout << endl << "(!) ATTENZIONE: il caricamento da file di nuovi dati causera' la "
      << endl << "               cancellazione dei dati correnti!" << endl << endl;
 
 /* Richiesta di prosecuzione */ 
 while (continua) {
   cout << "(?) PROSEGUIRE(s/n)?  ";
   fgets(buff, MAXSIZE, stdin);
   switch (buff[0]) {
     case 's':
       continua = 0;
       break;
     case 'n':
       return;            
   };     
 }
 continua = 1;
 
 /* Richiede nomefile da aprire e tenta di aprirlo */
 while(continua) {
   cout << endl;
   cout << "--> Inserisci il nome del file da cui caricare i dati(q per uscire): ";
   fgets(buff, MAXSIZE, stdin);
   buff[strlen(buff)-1] = 0;
   if (!strcmp(buff, "q") || !strcmp(buff, "Q"))
     return;
   if((fptr = fopen(buff, "r")) == NULL) 
      cout << endl << ":-( File inesistente o non apribile!" << endl;
   else   
      continua = 0;
 }
 
 /* Svuotamento dei dati da FH1 */
 FH1->reset();
 
 /* Finche' non viene trovata una linea priva di un numero, si inseriscono 
  * nuovi elementi in FH1
  */
 while(1) {
  fgets(buff, MAXSIZE, fptr);
  if (strlen(buff)==1 || feof(fptr))
    break;
  else
    buff[strlen(buff)-1] = 0; 
  num = strtok(buff, " \t\r");
  chiave = strtol(num, endptr, 10);
  if (**endptr == '\0') 
    FH1->insert(chiave);
 }
 
 /* Stamapa info dello heap caricato da file */
 cout << "\n************** HEAP CARICATO DA FILE **************";
 gestioneMIN(); 
 fclose(fptr);
}

/*************************** GESTIONE SALVATAGGIO SU FILE *****************************/
/* Salva su file tutte le chiavi dello heap corrente */

void gestioneSAVE() {
 int continua = 1;
 Elem *minimo;
 FILE* fptr; 
 ProgressBar pb = ProgressBar("SALVATAGGIO:", FH1->getNum_nodi(), 12, 40,'#');
  
 /* Stampa warning */ 
 cout << endl << "(!) ATTENZIONE: il salvataggio su file comporta la cancellazione "
      << endl << "                del contenuto di quel file!" << endl << endl;
 /* Richiesta di prosecuzione */
 while (continua) {
   cout << "(?) PROSEGUIRE(s/n)?  ";
   fgets(buff, MAXSIZE, stdin);
   switch (buff[0]) {
     case 's':
       continua = 0;
       break;
     case 'n':
       return;            
   };     
 }
 continua = 1;
 /* Richiede nomefile da aprire (se non esiste lo crea) */
 while(continua) {
   cout << endl;
   cout << "--> Inserisci il nome del file su cui salvare lo heap(q per uscire): ";
   fgets(buff, MAXSIZE, stdin);
   buff[strlen(buff)-1] = 0;
   if (!strcmp(buff, "q") || !strcmp(buff, "Q"))
     return;
   if((fptr = fopen(buff, "w")) == NULL) 
      cout << endl << ":-( Imposssibile aprire il file!" << endl;
   else   
      continua = 0;
 } 
 
 /* Continua a estrarre il minimo, salvare la sua chiave nel file e inserirlo in FH2 */
 while(FH1->getNum_nodi()  != 0) {
   pb.Advance();
   minimo = FH1->extractMin();
   FH2->insert(minimo->getChiave());
   fprintf(fptr, "%d\n", minimo->getChiave());
 }
 
 /* Libera la memoria puntata da FH1 */ 
 delete(FH1); 
 /* FH1 diventa FH2 e viene allocata nuova memoria per FH2 */
 FH1 = FH2;
 FH2 = new FibHeap();
 /* Chiusura file e stampa msg */
 fclose(fptr);
 cout << ":-) Salvataggio avvenuto correttamente!" << endl;
}

/**************** GESTIONE UNIONE DELLO HEAP CON UNO CARICATO DA FILE *****************/
/* Chima la unify() di Fh1 con FH2 che un nuovo heap caricato da file */ 

void gestioneUNION() {
int continua = 1;
 FILE *fptr;
 long int chiave;
 char **endptr = (char**)malloc(sizeof(char*));
 char *num = (char*)malloc(MAXSIZE*sizeof(char));

 /* Richiede nomefile da aprire e tenta di aprirlo */
 while(continua) {
   cout << endl;
   cout << "--> Inserisci il nome del file da cui caricare i dati per l'unione(q per uscire): ";
   fgets(buff, MAXSIZE, stdin);
   buff[strlen(buff)-1] = 0;
   if (!strcmp(buff, "q") || !strcmp(buff, "Q"))
     return;
   if((fptr = fopen(buff, "r")) < 0) 
      cout << endl << ":-( File inesistente o non apribile!" << endl;
   else   
      continua = 0;
 }
 
 /* Svuoto FH2 dai vecchi dati */
 FH2->reset();
 
 /* Legge ogni riga del file e compie l'inserimento in FH2 */
 while(1) {
  fgets(buff, MAXSIZE, fptr);
  if (strlen(buff)==1 || feof(fptr))
    break;
  else
    buff[strlen(buff)-1] = 0; 
  num = strtok(buff, " \t\r");
  chiave = strtol(num, endptr, 10);
  if (**endptr == '\0')
    FH2->insert(chiave);
} 
  
 /* Stampa info su FH1 prima e dopo l'unione con FH2 */
 cout << endl << "**************** PRIMA DELL'UNIONE *******************";
 gestioneMIN();
 FH1->unify(&FH2);
 cout << endl << "****************** DOPO L'UNIONE *********************"; 
 gestioneMIN();

 /* Chiusura del file */
 fclose(fptr);
}
