/************************************************
**** FIBHEAP:                                ****
**** classe che implementa gli heap          ****
**** di Fibonacci                            ****
****                                         ****
**** Larcheri Paolo       52 SI              ****
************************************************/
#include <math.h>
#include <string.h>
#include <iostream.h>
#include "ProgressBar.h"

#ifndef ELEM
  #include "Elem.h"
#endif
 
#ifndef  N_N
  #include "Nodo_Nodo.h"
#endif

template <class T>
class FibHeap {

  private:
  
    Elem<T> *testa;                /* puntatore alla prima radice */
    Elem<T> *min;                  /* puntatore all'elemento con chiave minima */ 
    
    int num_nodi;              
    int num_alberi;
    int num_marcati;
    
    void setTesta(Elem<T> *t) {
        this->testa = t;
    }

    void FibHeap::setMin(Elem<T> *m) {
        this->testa = m;
    }

    void setNum_nodi(int nn) {
        this->num_nodi = nn;
    }

    void setNum_alberi(int na) {
        this->num_alberi = na;
    }

    void setNum_marcati(int nm) {
        this->num_marcati = nm;
    }

    
   void consolidate() {
     /* NB: se A[i]=y allora y e' una radice con grado i */
     Elem<T> **A, *nodo_curr1, *nodo_curr2, *deposito, *next;
     int i, degree, aut1aut2, fine;
  
     /* inizializzo A[] a NULL */
     A = (Elem<T>**)malloc(((this->num_nodi)+1)*sizeof(Elem<T>*));
     for(i=0; i<(this->num_nodi); i++) 
        A[i] = NULL;
  
     next = this->min;
     fine = this->num_alberi;
  
     /* per ogni radice dello heap... */
     for (i=1; i<=fine; i++) {  
       nodo_curr1 = next;
       next = next->getDestro();
       degree = nodo_curr1->getGrado();
       /* finche' c'e' un altra radice con lo stesso grado... */
       while(A[degree] != NULL) {
         nodo_curr2 = A[degree];
         if ((nodo_curr1->getChiave()) > (nodo_curr2->getChiave())) {
	   deposito = nodo_curr1;
           nodo_curr1 = nodo_curr2;
           nodo_curr2 = deposito;
         }
         /* fondo i due alberi in unico albero di grado (degree+1) */
         this->link(nodo_curr2, nodo_curr1);
         A[degree] = NULL;
         degree++; 
       }
    
       A[degree] = nodo_curr1;
    }
  
     /* metto MOMENTANEAMENTE a NULL la testa e il minimo */
     (this->testa) = NULL;
     (this->min) = NULL;
  
     /* ricompongo la lista delle radici dello heap con tutti gli elementi contenuti in A[] */
     for (i=0; i<(this->num_nodi); i++) {
       if(A[i] != NULL) {
         if((this->testa) == NULL) {
           A[i]->setDestro(A[i]);
           A[i]->setSinistro(A[i]);
         }
         else {
     	   A[i]->setSinistro((this->testa)->getSinistro());
	   ((this->testa)->getSinistro())->setDestro(A[i]);
	   (this->testa)->setSinistro(A[i]);
           A[i]->setDestro(this->testa);
         }
      
         (this->testa) = A[i];
      
         /* se necessario aggiorno il minimo */
         if((this->min == NULL) || ((this->min)->getChiave() > A[i]->getChiave())) 
           this->min = A[i];
       }
     }
  
     /* libero la a cui punta A[]*/
     free(A);
   }


   void link(Elem<T> *T1, Elem<T> *T2) {
     Elem<T> *figlio;
  
     /* Tolgo T1 dalla lista delle radici */
     (T1->getSinistro())->setDestro(T1->getDestro());
     (T1->getDestro())->setSinistro(T1->getSinistro());

     /* T1 diventa PRIMO figlio di T2 */  
     T1->setPadre(T2);
     T2->setGrado(T2->getGrado() + 1);
     this->setNum_alberi(this->getNum_alberi() - 1);
     if((figlio=T2->getFigli()) != NULL) {
       T1->setDestro(figlio);
       T1->setSinistro(figlio->getSinistro());
       (figlio->getSinistro())->setDestro(T1);
       figlio->setSinistro(T1);
     }
     else{
       T1->setDestro(T1);
       T1->setSinistro(T1);
     }
     T2->setFigli(T1);
   }


   void cut(Elem<T>* fi, Elem<T>* pa) {
  
     /* tolgo il figlio dalla lista dei figli */
     if (pa->getGrado() == 1) 
       pa->setFigli(NULL);
     else {
       pa->setFigli(fi->getDestro());
       (fi->getDestro())->setSinistro(fi->getSinistro());
       (fi->getSinistro())->setDestro(fi->getDestro());
     }
  
     /* il figlio non ha padre e il padre ha un figlio in meno */
     fi->setPadre(NULL);
     pa->setGrado(pa->getGrado() - 1);
  
     /* aggiungo il figlio in testa alla lista delle radici */
     fi->setDestro(this->testa);
     fi->setSinistro(this->testa->getSinistro());
     (this->testa->getSinistro())->setDestro(fi);
     (this->testa)->setSinistro(fi);
     this->testa = fi;
  
     fi->unmark(); 
     num_marcati--;
     num_alberi++;
   }


   void cascadingCut(Elem<T> *curr_nodo){
     Elem <T>*padre;
  
     padre = curr_nodo->getPadre();
     /* tramite chiamate ricorsive risalgo l'albero finche' non trovo un padre non marcato
        (cha verra' marcato), altrimenti lo promuovo a radice e proseguo nella risalita */
     if(padre != NULL) {
       if(!(curr_nodo->isMarked()))
         curr_nodo->mark();
       else {
         this->cut(curr_nodo, padre);   
         this->cascadingCut(padre);
       }
     }
   }

    
  public:
  
    
   FibHeap() {
     /* inizializzazione degli attributi */
     this->testa = NULL;
     this->min = NULL;
     this->num_nodi = 0;
     this->num_alberi = 0;
   }

   ~FibHeap() {
   }
 
   
   
   Elem<T>* getTesta() {
       return this->testa;
    }

    int FibHeap::getNum_nodi() {
       return this->num_nodi;
    }

    int FibHeap::getNum_alberi() {
       return this->num_alberi;
    }

    int FibHeap::getNum_marcati() {
       return this->num_marcati;
    }
    
    Elem<T>* getMin() {
       return this->min;
    }

    
    Elem<T>* insert(int n1, int n2, T chiave) {
      Elem<T> *newElem = new Elem<T>(chiave, n1, n2);
  
      /* aggiungo l'elemento in testa alla lista delle radici */
      if((this->testa) != NULL) {
        /* se lo heap e' vuoto... */
        newElem->setDestro(this->testa);
        newElem->setSinistro((this->testa)->getSinistro());
        (testa->getSinistro())->setDestro(newElem);
        (this->testa)->setSinistro(newElem);
      }
      else {
        /* se lo heap non e' vuoto... */
        newElem->setDestro(newElem);
        newElem->setSinistro(newElem);
      }
      testa = newElem;
  
      /* eventualmente aggiorno il 'puntatore al minimo' dello heap */
      if ((this->min == NULL) || (min->getChiave() >= newElem->getChiave()))
        (this->min) = newElem;
  
      /* incremento il numero di nodi e di radici */  
      (this->num_alberi)++;
      (this->num_nodi)++;
 
      return newElem; 
    }


   void unify(FibHeap **FH2) {
     Elem<T> *sin1;
     Elem<T> *sin2;
  
     /* se lo heap e' vouto, questo diventa praticamente quello di cui 
        e' passato l'indirizzo */
     if (this->getTesta() == NULL) {
       this->testa = (*FH2)->getTesta();
       this->min = (*FH2)->getMin();
       delete(*FH2);
       (*FH2) = new FibHeap();
       return;
     }
     /* se lo heap  di cui e' passato l'indirizzo e' vuoto, ritorno*/
     if ((*FH2)->getTesta() == NULL) {
       return;
     }
  
     /* appendo le radici di uno alla lista delle radici dell'altro */
     sin1 = (this->testa)->getSinistro();
     sin2 = ((*FH2)->getTesta())->getSinistro();
     sin1->setDestro((*FH2)->getTesta());
     sin2->setDestro(this->testa);
     ((*FH2)->getTesta())->setSinistro(sin1);
     (this->testa)->setSinistro(sin2);
  
     /* aggiorno gli attributi/contatori */
     this->setNum_nodi(this->getNum_nodi() + (*FH2)->getNum_nodi());  
     this->setNum_alberi(this->getNum_alberi() + (*FH2)->getNum_alberi());  
     this->setNum_marcati(this->getNum_marcati() + (*FH2)->getNum_marcati());  
  
     /* se serve aggiorno il puntatore al minimo */
     if((this->getMin())->getChiave()  >  ((*FH2)->getMin())->getChiave())
        (this->min) = ((*FH2)->getMin());
  
     /* libero la memoria occupata dallo heap passato per indirizzo
        e ne alloco dell'altra */   
     delete(*FH2);
     *FH2 = new FibHeap<T>();
   }



   Elem<T>* extractMin() {
     Elem<T> *figlio;  
     Elem<T> *ultimo_figlio;
     Elem<T> *minimo;
     int num_figli;
  
     minimo = NULL;
    
     if((this->min) != NULL) {
       minimo = this->min;
       num_figli = minimo->getGrado();

       /* aggiungo alla lista delle radici tutti i figli di this->min */
       if((figlio = minimo->getFigli()) != NULL) {
          ultimo_figlio = figlio->getSinistro();
         /* tutti gli ex-figli di this->min non hanno piu' padre */
         do {
           figlio->setPadre(NULL);
	   figlio = figlio->getDestro();    
         } while(figlio != (ultimo_figlio->getDestro()));

         ultimo_figlio = figlio->getSinistro();
         figlio->setSinistro(minimo);
         ultimo_figlio->setDestro(minimo->getDestro());
         (minimo->getDestro())->setSinistro(ultimo_figlio);
         minimo->setDestro(figlio);
     
       }
    
       /* aggiorno gli attributi/contatori */
       (this->num_nodi)--;
       (this->num_alberi) += (num_figli-1);
        
       /* tolgo il minimo dalla lista della radici */
       if((this->num_nodi) > 0) {
         (minimo->getSinistro())->setDestro(minimo->getDestro());
         (minimo->getDestro())->setSinistro(minimo->getSinistro());
       }
 
       /* aggiorno il minimo e la testa dello heap */
       if((minimo->getDestro()) == minimo) {
         this->min = NULL;
         this->testa = NULL;
       }
       else {
         if(this->testa == this->min)
           this->testa = minimo->getDestro();
         this->min = minimo->getDestro();
         this->consolidate();    /* RIMETTE A POSTO LO HEAP */
       }
     }
  
     /* ritorno il nodo con CHIAVE minimo */
     /* NOTA: sara' il chiamante a dover deallocare la memoria da lui occupata */
     return minimo;
   }




   int decreaseKey(Elem<T>* arco, int nd, int ps) {
     Elem<T> *arco_padre;
    
    if ((arco == NULL) || (arco->getChiave() <= ps))
       return (0);
  
     arco->setChiave(ps);
     arco->setNodo1(nd);
     arco_padre = arco->getPadre();
  
     if((arco_padre != NULL) && (arco->getChiave() < arco_padre->getChiave())) {
       cut(arco, arco_padre);
       cascadingCut(arco_padre);
     }
  
     if(arco->getChiave() < min->getChiave())
       min = arco;
  
     return (1);
   }


   void reset() {
     Elem<T> *minimo;
     ProgressBar pb = ProgressBar("DEALLOCAZIONE MEMORIA:", this->getNum_nodi(), 12, 40,'#');
  
     while(this->testa != NULL){
       /* per ogni nodo dello heap viene liberata la memoria */
       minimo = this->extractMin();
       delete(minimo);
       pb.Advance();
     }
   }
     
      
};






/*********************** INSERISCE UN NUOVO ELEMENTO NELLO HEAP ***********************/
/* prende in input un numero e aggiunge allo heap un nuovo elemento con quella Peso */

