#include "Grafo.h"
#include <stdlib.h>
#include <iostream.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
 
#define MAXLEN 20
#define MAXEDGEVALUE 100.0  


/**************************** COSTRUTTORE E DISTRUTTORE*****************************/

Grafo::Grafo() {
  num_nodi = 0;
  num_archi = 0;
  max_num_nodi = MAX_NUM_NODES;
  adiac_list = new list<Nodo_Nodo*>[max_num_nodi];
}  

Grafo::~Grafo() {
  delete[] adiac_list;
}


/********************** RITORNANO I VARI ATTRIBUTI DEL GRAFO ***********************/

int Grafo::getMax_num_nodi() {
  return max_num_nodi;
}

int Grafo::getNum_archi() {
  return num_archi;
}

int Grafo::getNum_nodi() {
  return num_nodi;
}

list<Nodo_Nodo*>* Grafo::getAdiac_list() {
  return adiac_list;
}

/************** INSERISCE UN NUOVO ARCO (E QUINDI I NODI SE NECESS.) ***************/

int Grafo::insArco(int n1, int n2, int p) {
  Nodo_Nodo *newarco1;
  Nodo_Nodo *newarco2;
  
  if((n1 == n2) || (n1 < 0) || (n2 < 0) || (p == 0)
      || (n1 >= max_num_nodi) || (n2 >= max_num_nodi))
    return 0;
  if(esisteArco(n1, n2))
    return 0;  
  
  newarco1 = new Nodo_Nodo(n2, p);
  newarco2 = new Nodo_Nodo(n1, p);
  
  if(adiac_list[n1].empty())
    num_nodi++;
  if(adiac_list[n2].empty())
    num_nodi++;
  num_archi++;  
  
  adiac_list[n1].push_back(newarco1);
  adiac_list[n2].push_back(newarco2);
  
  return 1;
}


/*********************** VERIFICA SE UN ARCO APPARTIENE AL GRAFO ********************/

bool Grafo::esisteArco(int i,int j) {
  bool ret = false;
  list< Nodo_Nodo* >::iterator it;
  
  for(it=adiac_list[i].begin(); it!=adiac_list[i].end(); it++) 
    if((*it)->getNodo() == j)
      ret = true;
  return ret;
}


/*********************** VERIFICA SE UN NODO APPARTIENE AL GRAFO ********************/

bool Grafo::esisteNodo(int nd) {
  return (!adiac_list[nd].empty());
}


/************************ RITORNA IL PESO DELL'ARCO RICHIESTO ***********************/

int Grafo::getPesoArco(int i,int j) {
  int ret = -1;
  list<Nodo_Nodo*>::iterator it;
  
  for(it=adiac_list[i].begin(); it!=adiac_list[i].end(); it++) 
    if((*it)->getNodo() == j)
      ret = (*it)->getPeso();
  return ret;
}


/************** LEGGE UN GRAFO DA FILE (vengono persi i vecchi dati) ****************/

int Grafo::loadFile(char* nf) {
  FILE *fp;
  int n1, n2, p, i;
  char buff[MAXLEN];
  char *num;
  
  if((fp = fopen(nf, "r")) == NULL)
    return 0;
  
  clear();  /* pulisce il grafo */
    
  while(1) {
    fgets(buff, MAXLEN, fp);
    if (strlen(buff)==1 || feof(fp))
      break;
    else
      buff[strlen(buff)-1] = 0;
    num = strtok(buff, "-");
    n1 = atoi(num);
    num = strtok(NULL, " \t");
    n2 = atoi(num);
    num = strtok(NULL, "\r\n");
    p = atoi(num);
    insArco(n1, n2, p);
  }
  fclose(fp);
     
  printf("NUMNODI = %d\n", num_nodi);
  printf("NUMARCHI = %d\n", num_archi);
  return 1;
}

/**************************** SALVA SU FILE I DATI DEL GRAFO *************************/

int Grafo::saveFile(char* nf){
  FILE* fp;
  int nd, i, j;
  list<Nodo_Nodo*>::iterator it;
  
  fp = fopen(nf, "w");
  if(!fp)
    return 0;
  
  for (i=0; i < max_num_nodi; i++)
    for(it=adiac_list[i].begin(); it!=adiac_list[i].end(); it++) 
      if((*it)->getNodo() > i)
        fprintf(fp, "%d-%d %d\n", i, (*it)->getNodo(), (*it)->getPeso());  
  
  fclose(fp);
  return 1;
}

/**************************** SALVA SU FILE I DATI DEL GRAFO *************************/
/****************************** (visualizzabile con Dot) ***************************/

int Grafo::saveDotFile(char* nf, Grafo* T, bool* N, Elem** FHtracer){
  FILE* fp;
  int i;
  char comando[80];
  list<Nodo_Nodo*>::iterator it;
  
  fp = fopen(nf, "w");
    
  if(!fp) {
    printf(":( Impossibile salvare su file temporaneo!\n");
    return 0;
    
  }
  
  fprintf(fp, "graph G {\n");
  fprintf(fp, "size=\"7.5,10\"; ratio=compress; center=false\n");
  fprintf(fp, "node [shape=ellipse, height=.05, width=.1, fontsize=12];\n");
  for (i=0; i<max_num_nodi; i++) {
    if (!adiac_list[i].empty()) {
      fprintf(fp, "%d ", i);
      if((T != NULL) && (N != NULL) && (N[i] == false))
        fprintf(fp, "[color=grey15, fontcolor=grey15];\n");
      else
        fprintf(fp, "[style=bold, color=black, fontcolor=black];\n");
    }
  }
  for (i=0; i < max_num_nodi; i++) {
    for(it=adiac_list[i].begin(); it!=adiac_list[i].end(); it++) {
      if((*it)->getNodo() > i) {
        if((T != NULL)  &&  (!T->esisteArco(i,(*it)->getNodo()))) {
          
	  if   ((FHtracer!=NULL) && ((FHtracer[(*it)->getNodo()] != NULL  &&   FHtracer[(*it)->getNodo()]->getNodo1() == i)
	       || (FHtracer[i] != NULL  &&   FHtracer[i]->getNodo1() == (*it)->getNodo())))
	    
	    fprintf(fp, "%d -- %d [label=\"%d\", len=2, fontsize=10, color=red, fontcolor=red, style=dashed]\n", i, (*it)->getNodo(), (*it)->getPeso());
	  
	  else
       	    
	    fprintf(fp, "%d -- %d [label=\"%d\", len=2, fontsize=10, color=grey15, fontcolor=grey15]\n", i, (*it)->getNodo(), (*it)->getPeso());
	}    
        
	else
	 
	  fprintf(fp, "%d -- %d [label=\"%d\", len=2, fontsize=10, color=black, style=bold, fontstyle=bold]\n", i, (*it)->getNodo(), (*it)->getPeso());  
     }
    }
  }
  
  fprintf(fp, "}");
  fclose(fp);
  
  sprintf(comando, "./dot -Tps -Gepsilon=0.001 %s -o %s.ps", nf, nf);
  system(comando);
  
  sprintf(comando, "pstogif -out %s.gif -depth 8 -density 120 %s.ps ", nf,
  nf);//>& ziocane.txt
  system(comando);
  
  sprintf(comando, "rm %s.ps ", nf);
  system(comando);
  
  sprintf(comando, "rm %s", nf);
  system(comando);
  
  return 1;
}

//>& /dev/null
/************************** GENERA UN GRAFO CASUALE (CONNESSO) ***********************/

int Grafo::randCreate(int nn, int na) {
  int i, j, nodo1, nodo2, ps;
  bool non_occupato;
  
  if ((na < (nn-1)) || (na > ((nn * (nn-1))/2))) {
    cout << ":( Impossibile creare il grafo -> numero di archi e nodi incompatibili!\n";
    return 0;
  }
    
  clear(); /* pulisce il grafo */
 
  num_archi = na;
  num_nodi = 0;
  
  srand(time(NULL));
  
  for (i=0; i<(nn-1); i++) {
    do {
     j = (i+1) + rand()%(nn-i-1);
       }  while(esisteArco(i, j));	
    
    ps = 1 + (int) (MAXEDGEVALUE * rand()/ (RAND_MAX+1.0));
    insArco(i, j, ps);
  }
    
  for (i=nn-1; i<na; i++) {
    do {
         nodo1 = rand()% nn;
         nodo2 = nodo1;
	 while(nodo2 == nodo1)
           nodo2 = rand()% nn;
       }  while (esisteArco(nodo1, nodo2));
    ps = 1 + (int) (MAXEDGEVALUE * rand()/ (RAND_MAX+1.0));
    insArco(nodo1, nodo2, ps);  
  }
  cout << ":) Grafo creato correttamente!\n";  
  return 1; 
}


/************************** SVUOTA IL GRAFO DAI VECCHI DATI **************************/

void Grafo::clear() {
  /* TOLGO I VECCHI DATI DAL GRAFO */
  delete[] adiac_list;
  num_nodi = 0;
  num_archi = 0;
  adiac_list = new list<Nodo_Nodo*>[max_num_nodi];
}
